"""
********************************************************************************
compas_eve
********************************************************************************

.. currentmodule:: compas_eve


Classes
=======

.. autosummary::
    :toctree: generated/
    :nosignatures:

    Message
    Topic
    Publisher
    Subscriber
    Transport
    InMemoryTransport
    get_default_transport
    set_default_transport

"""

from __future__ import print_function

import os


__author__ = ["Gonzalo Casas"]
__copyright__ = "Gramazio Kohler Research"
__license__ = "MIT License"
__email__ = "casas@arch.ethz.ch"
__version__ = "0.3.5"

from .event_emitter import EventEmitterMixin  # noqa: F401 needed here to avoid circular import on py2.7
from .core import Message, Publisher, Subscriber, Transport, Topic, get_default_transport, set_default_transport
from .memory import InMemoryTransport

HERE = os.path.dirname(__file__)
HOME = os.path.abspath(os.path.join(HERE, "../../"))

set_default_transport(InMemoryTransport())

__all__ = [
    "HOME",
    "Message",
    "Publisher",
    "Subscriber",
    "Topic",
    "Transport",
    "get_default_transport",
    "set_default_transport",
    "InMemoryTransport",
]
__all_plugins__ = ["compas_eve.rhino.install"]



"""
Background Task component.

Executes a long-running task in the background, while keeping Grasshopper reactive.

COMPAS EVE v0.3.5
"""

import threading
import scriptcontext as sc

from compas_eve.ghpython import BackgroundWorker
from ghpythonlib.componentbase import executingcomponent as component


DEBUG = False


class BackgroundTaskComponent(component):
    def RunScript(self, reset, task, on):
        if not on:
            BackgroundWorker.stop_instance_by_component(ghenv)  # noqa: F821
            return None

        self.worker = BackgroundWorker.instance_by_component(ghenv, task, force_new=reset)  # noqa: F821

        if not self.worker.is_working() and not self.worker.is_done() and reset:
            self.worker.start_work()

        print("Worker ID: {}".format(id(self.worker)))
        print("Is worker running? {}".format(self.worker.is_working()))
        print("Worker completed? {}".format(self.worker.is_done()))
        if hasattr(self.worker, "thread"):
            print("Worker thread: {}".format(self.worker.thread))

        if DEBUG:
            other_workers = []
            for key in sc.sticky.keys():
                if key.startswith("background_worker_"):
                    worker = sc.sticky[key]
                    if worker != self.worker:
                        other_workers.append(worker)

            if len(other_workers):
                print
                print("Found {} more workers:".format(len(other_workers)))
                for worker in other_workers:
                    print("* Worker ID: {}".format(id(worker)))
                    print(" - Is worker running? {}".format(worker.is_working()))
                    print(" - Worker completed? {}".format(worker.is_done()))
                    print(" - Worker thread: {}".format(worker.thread))

            non_main_threads = [thread for thread in threading.enumerate() if thread.name != "MainThread"]
            if len(non_main_threads):
                print
                print("Found {} background threads running:".format(len(non_main_threads)))
                for thread in non_main_threads:
                    print(" - " + thread.name)

        if hasattr(self.worker, "result"):
            return self.worker.result
        else:
            return None



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import threading

import Rhino
import scriptcontext
import System
from compas_ghpython import create_id
from compas_ghpython import update_component


class BackgroundWorker(object):
    """Background worker simplifies the creation of long-running tasks inside Grasshopper.

    A long-running task is any piece of code that will run for an extended period of time,
    for example, a very complex calculation, or loading a very large data file, etc.

    To use it, write your long-running function in a Grasshopper GHPython component,
    and pass it as the input to the background worker component.
    The worker will continue working without blocking the UI.

    The following is an example of a long-running function that updates the
    progress while it runs.

    .. code-block:: python

        import time

        def do_something_long_and_complicated(worker):
            # Result can be of any data type
            result = 0

            for i in range(50):
                worker.current_value = i
                result += i
                worker.display_progress(i / (50-1))
                time.sleep(0.01)

            worker.display_message("Done!")

            return result


    Parameters
    ----------
    ghenv : ``GhPython.Component.PythonEnvironment``
        Grasshopper environment object
    long_running_function : function, optional
        This function will be the main entry point for the long-running task.
    """

    def __init__(self, ghenv, long_running_function=None):
        super(BackgroundWorker, self).__init__()
        self.ghenv = ghenv
        self._is_working = False
        self._is_done = False
        self._is_cancelled = False
        self._has_requested_cancellation = False
        self.long_running_function = long_running_function

    def is_working(self):
        """Indicate whether the worker is currently working or not."""
        return self._is_working

    def is_done(self):
        """Indicate whether the worker is done or not."""
        return self._is_done

    def has_requested_cancellation(self):
        return self._has_requested_cancellation

    def request_cancellation(self):
        """Mark the current worker as cancelled, so that the background task can stop processing."""
        self._has_requested_cancellation = True

    def start_work(self):
        """Start the background processing thread where work will be performed."""

        def _long_running_task_wrapper(worker):
            try:
                worker.set_internal_state_to_working()
                result = self.long_running_function(self)
                worker.set_internal_state_to_done(result)
            except Exception as e:
                worker.display_message(str(e))
                worker.set_internal_state_to_cancelled()

        target = _long_running_task_wrapper
        args = (self,)
        self.thread = threading.Thread(target=target, args=args)
        self.thread.start()

    def set_internal_state_to_working(self):
        """Set the internal state to ``working``."""
        self._is_working = True
        self._is_done = False
        self._is_cancelled = False

    def set_internal_state_to_done(self, result):
        """Set the internal state to ``done``, which indicates the worker has completed."""
        self._is_working = False
        self._is_done = True
        self._is_cancelled = False
        self.update_result(result, delay=1)

    def update_result(self, result, delay=1):
        """Update the result of the worker.

        This will update the result of the worker, and trigger a solution expiration
        of the Grasshopper component.

        Parameters
        ----------
        result : object
            Result of the worker.
        delay : int, optional
            Delay (in milliseconds) before updating the component, by default 1.
        """
        self.result = result
        update_component(self.ghenv, delay)

    def set_internal_state_to_cancelled(self):
        """Set the internal state to ``cancelled``."""
        self._is_working = False
        self._is_done = False
        self._is_cancelled = True

    def display_progress(self, progress):
        """Display a progress indicator in the component.

        Parameters
        ----------
        progress : float
            Float between ``0..1`` indicating progress of completion.
        """
        self.display_message("Progress {:.1f}%".format(progress * 100))

    def display_message(self, message):
        """Display a message in the component without triggering a solution expiration.

        Parameters
        ----------
        message : str
            Message to display.

        """

        def ui_callback():
            self.ghenv.Component.Message = message
            self.ghenv.Component.OnDisplayExpired(True)

        Rhino.RhinoApp.InvokeOnUiThread(System.Action(ui_callback))

    @classmethod
    def instance_by_component(cls, ghenv, long_running_function=None, force_new=False):
        """Get the worker instance assigned to the component.

        This will get a persistant instance of a background worker
        for a given component. The parameter `force_new` can
        be set to `True` to request a new instance to be created.

        Parameters
        ----------
        ghenv : ``GhPython.Component.PythonEnvironment``
            Grasshopper environment object
        long_running_function : function, optional
            This function will be the main entry point for the long-running task.
        force_new : bool, optional
            Force the creation of a new background worker, by default False.

        Returns
        -------
        :class:`BackgroundWorker`
            Instance of the background worker of the current component.
        """

        key = create_id(ghenv.Component, "background_worker")
        worker = scriptcontext.sticky.get(key)

        if worker and force_new:
            worker.request_cancellation()
            worker = None
            del scriptcontext.sticky[key]

        if not worker:
            worker = cls(ghenv, long_running_function=long_running_function)
            scriptcontext.sticky[key] = worker

        return worker

    @classmethod
    def stop_instance_by_component(cls, ghenv):
        """Stops the worker instance assigned to the component.

        If there is no worker running, it will do nothing.

        Parameters
        ----------
        ghenv : ``GhPython.Component.PythonEnvironment``
            Grasshopper environment object
        """

        key = create_id(ghenv.Component, "background_worker")
        worker = scriptcontext.sticky.get(key)

        if worker:
            worker.request_cancellation()
            worker = None
            del scriptcontext.sticky[key]



# -*- coding: utf-8 -*-

# The following code is derivative work of [pyee](https://github.com/jfhbrook/pyee)
# Copyrighted by Joshua Holbrook with a MIT license


# The MIT License(MIT)

# Copyright(c) 2015 Joshua Holbrook

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files(the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and / or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
``EventEmitterMixin`` is similar to the ``EventEmitter`` of Node.js.
It supports both synchronous callbacks and asyncio coroutines.
"""

try:
    from asyncio import ensure_future
    from asyncio import iscoroutine
except ImportError:
    iscoroutine = None
    ensure_future = None

from collections import OrderedDict
from collections import defaultdict
from threading import RLock

__all__ = ["EventEmitterMixin", "EventEmitterException"]


class EventEmitterException(Exception):
    """An internal exception."""

    pass


class EventEmitterMixin(object):
    """Mixin to add event emitter features to a class.

    For interoperation with asyncio, one can specify the scheduler and
    the event loop. The scheduler defaults to ``asyncio.ensure_future``,
    and the loop defaults to ``None``. When used with the default scheduler,
    this will schedule the coroutine onto asyncio's default loop.

    This should also be compatible with recent versions of twisted by
    setting ``scheduler=twisted.internet.defer.ensureDeferred``.

    Most events are registered with EventEmitterMixin via the ``on`` and ``once``
    methods. However, event emitters have two *special* events:

    - ``new_listener``: Fires whenever a new listener is created. Listeners for
      this event do not fire upon their own creation.

    - ``error``: When emitted raises an Exception by default, behavior can be
      overriden by attaching callback to the event.

      For example::

          @ee.on('error')
          def onError(message):
              logging.err(message)

          ee.emit('error', Exception('something blew up'))

      For synchronous callbacks, exceptions are **not** handled for you---
      you must catch your own exceptions inside synchronous ``on`` handlers.
      However, when wrapping **async** functions, errors will be intercepted
      and emitted under the ``error`` event. **This behavior for async
      functions is inconsistent with node.js**, which unlike this package has
      no facilities for handling returned Promises from handlers.
    """

    def __init__(self, *args, **kwargs):
        super(EventEmitterMixin, self).__init__(*args, **kwargs)
        self._events = defaultdict(OrderedDict)
        self._schedule = kwargs.get("scheduler", ensure_future)
        self._loop = kwargs.get("loop", None)
        self._event_lock = RLock()

    def on(self, event, f=None):
        """Registers the function (or optionally an asyncio coroutine function)
        ``f`` to the event name ``event``.

        If ``f`` isn't provided, this method returns a function that
        takes ``f`` as a callback; in other words, you can use this method
        as a decorator, like so::

            @ee.on('data')
            def data_handler(data):
                print(data)

        As mentioned, this method can also take an asyncio coroutine function::

           @ee.on('data')
           async def data_handler(data)
               await do_async_thing(data)


        This will automatically schedule the coroutine using the configured
        scheduling function (defaults to ``asyncio.ensure_future``) and the
        configured event loop (defaults to ``asyncio.get_event_loop()``).

        In both the decorated and undecorated forms, the event handler is
        returned. The upshot of this is that you can call decorated handlers
        directly, as well as use them in remove_listener calls.
        """

        with self._event_lock:

            def _on(f):
                self._add_event_handler(event, f, f)
                return f

            if f is None:
                return _on
            else:
                return _on(f)

    def _add_event_handler(self, event, k, v):
        # Fire 'new_listener' *before* adding the new listener!
        self.emit("new_listener", event, k)

        # Add the necessary function
        # Note that k and v are the same for `on` handlers, but
        # different for `once` handlers, where v is a wrapped version
        # of k which removes itself before calling k
        self._events[event][k] = v

    def emit(self, event, *args, **kwargs):
        """Emit ``event``, passing ``*args`` and ``**kwargs`` to each attached
        function. Returns ``True`` if any functions are attached to ``event``;
        otherwise returns ``False``.

        Example::

            ee.emit('data', '00101001')

        Assuming ``data`` is an attached function, this will call
        ``data('00101001')'``.

        For coroutine event handlers, calling emit is non-blocking. In other
        words, you do not have to await any results from emit, and the
        coroutine is scheduled in a fire-and-forget fashion.
        """
        handled = False

        with self._event_lock:
            for f in list(self._events[event].values()):
                result = f(*args, **kwargs)

                # If f was a coroutine function, we need to schedule it and
                # handle potential errors
                if iscoroutine and iscoroutine(result):
                    if self._loop:
                        d = self._schedule(result, loop=self._loop)
                    else:
                        d = self._schedule(result)

                    # scheduler gave us an asyncio Future
                    if hasattr(d, "add_done_callback"):

                        @d.add_done_callback
                        def _callback(f):
                            exc = f.exception()
                            if exc:
                                self.emit("error", exc)

                    # scheduler gave us a twisted Deferred
                    elif hasattr(d, "addErrback"):

                        @d.addErrback
                        def _callback(exc):
                            self.emit("error", exc)

                handled = True

        if not handled and event == "error":
            if args:
                raise args[0]
            else:
                raise EventEmitterException("Uncaught, unspecified 'error' event.")

        return handled

    def once(self, event, f=None):
        """The same as ``ee.on``, except that the listener is automatically
        removed after being called.
        """

        with self._event_lock:

            def _wrapper(f):
                def g(*args, **kwargs):
                    self.remove_listener(event, f)
                    # f may return a coroutine, so we need to return that
                    # result here so that emit can schedule it
                    return f(*args, **kwargs)

                self._add_event_handler(event, f, g)
                return f

            if f is None:
                return _wrapper
            else:
                return _wrapper(f)

    def off(self, event, f):
        """Removes the function ``f`` from ``event``."""
        self._events[event].pop(f)

    def remove_listener(self, event, f):
        """Removes the function ``f`` from ``event``."""
        self._events[event].pop(f)

    def remove_all_listeners(self, event=None):
        """Remove all listeners attached to ``event``.
        If ``event`` is ``None``, remove all listeners on all events.
        """
        with self._event_lock:
            if event is not None:
                self._events[event] = OrderedDict()
            else:
                self._events = defaultdict(OrderedDict)

    def listeners(self, event):
        """Returns a list of all listeners registered to the ``event``."""
        return list(self._events[event].keys())



# fmt: off
from __future__ import print_function

from ..core import Transport
from ..event_emitter import EventEmitterMixin

from compas.data import json_dumps
from compas.data import json_loads

import clr
import json
import os
import sys

lib_dir = os.path.join(os.path.dirname(__file__), "netlib")
if lib_dir not in sys.path:
    sys.path.append(lib_dir)

clr.AddReference("MQTTnet")

from System import Action
from System.Text import Encoding
from System.Threading import CancellationToken, CancellationTokenSource
from System.Threading.Tasks import Task

from MQTTnet import MqttFactory
from MQTTnet import MqttApplicationMessageBuilder
from MQTTnet.Client import MqttClientConnectResult
from MQTTnet.Client import MqttClientConnectResultCode
from MQTTnet.Client import MqttClientOptionsBuilder
from MQTTnet.Client import MqttClientDisconnectOptionsBuilder


class MqttTransport(Transport, EventEmitterMixin):
    """MQTT transport allows sending and receiving messages using an MQTT broker.

    Parameters
    ----------
    host : str
        Host name for the MQTT broker, e.g. ``broker.hivemq.com`` or ``localhost`` if
        you are running a local broker on your machine.
    port : int
        MQTT broker port, defaults to ``1883``.
    """
    def __init__(self, host, port=1883, *args, **kwargs):
        super(MqttTransport, self).__init__(*args, **kwargs)
        self.host = host
        self.port = port
        self._is_connected = False
        self._local_callbacks = {}

        self.cancellation_token_source = CancellationTokenSource()
        self.cancellation_token = self.cancellation_token_source.Token

        self.factory = MqttFactory()
        options = MqttClientOptionsBuilder().WithTcpServer(host, port).Build()

        self.client = self.factory.CreateMqttClient()
        self.client.ConnectAsync(options, self.cancellation_token).ContinueWith.Overloads[Action[Task[MqttClientConnectResult]]](self._on_connect)


    def _on_connect(self, event_args):
        if event_args.Result.ResultCode == MqttClientConnectResultCode.Success:
            self._is_connected = event_args.Result.ResultCode == MqttClientConnectResultCode.Success
            self.emit("ready")

    def close(self):
        options = MqttClientDisconnectOptionsBuilder().WithReasonString("Normal").Build()
        self.client.DisconnectAsync(options, CancellationToken.None) # noqa: E999 (disable flake8 error, which incorrectly parses None as the python keyword)

    def on_ready(self, callback):
        if self._is_connected:
            callback()
        else:
            self.once("ready", callback)

    def publish(self, topic, message):
        # TODO: can we avoid the additional cast to dict?
        application_message = (
            MqttApplicationMessageBuilder()
            .WithTopic(topic.name)
            .WithPayload(json_dumps(dict(message)))
            .Build()
        )

        def _callback(**kwargs):
            self.client.PublishAsync(application_message, CancellationToken.None)

        self.on_ready(_callback)

    def subscribe(self, topic, callback):
        event_key = "event:{}".format(topic.name)
        subscribe_id = "{}:{}".format(event_key, id(callback))

        def _local_callback(application_message):
            payload = Encoding.UTF8.GetString(application_message.Payload)
            msg = topic.message_type.parse(json_loads(payload))
            callback(msg)

        def _subscribe_callback(**kwargs):
            subscribe_opts = self.factory.CreateSubscribeOptionsBuilder().WithTopicFilter(lambda f: f.WithTopic(topic.name)).Build()
            self.client.ApplicationMessageReceivedAsync += self._on_message
            self.client.SubscribeAsync(subscribe_opts, self.cancellation_token)
            self.on(event_key, _local_callback)

        self._local_callbacks[subscribe_id] = _local_callback

        self.on_ready(_subscribe_callback)

        return subscribe_id

    def _on_message(self, event_args):
        event_key = "event:{}".format(event_args.ApplicationMessage.Topic)
        self.emit(event_key, event_args.ApplicationMessage)

    def advertise(self, topic):
        advertise_id = "advertise:{}:{}".format(topic.name, self.id_counter)
        # mqtt does not need anything here
        return advertise_id

    def unadvertise(self, topic):
        pass

    def unsubscribe_by_id(self, subscribe_id):
        ev_type, topic_name, _callback_id = subscribe_id.split(":")
        event_key = "{}:{}".format(ev_type, topic_name)

        callback = self._local_callbacks[subscribe_id]
        self.off(event_key, callback)
        del self._local_callbacks[subscribe_id]

        unsubscribe_opts = self.factory.CreateUnsubscribeOptionsBuilder().WithTopicFilter(topic_name).Build()
        self.client.UnsubscribeAsync(unsubscribe_opts, self.cancellation_token)
        self.client.ApplicationMessageReceivedAsync -= self._on_message

    def unsubscribe(self, topic):
        unsubscribe_opts = self.factory.CreateUnsubscribeOptionsBuilder().WithTopicFilter(topic.name).Build()
        self.client.UnsubscribeAsync(unsubscribe_opts, self.cancellation_token)
        self.client.ApplicationMessageReceivedAsync -= self._on_message
